"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

# These need to be here (or somewhere)
# to get the implementations added to the lists
import pyDE1.scale.acaia
import pyDE1.scale.atomax_skale_ii
import pyDE1.scale.felicita_arc
