# DO NOT EDIT -- This file automatically generated -- DO NOT EDIT
git_data = {
    "hash": "174df9c7",
    "subject": "WIP: Doc updates",
    "ref_names": "HEAD -> alpha-2023-01",
    "author_date": "Thu Dec 15 12:09:22 2022 -0800",
    "commit_date": "Sun Jan 22 17:19:48 2023 -0800",
}
source_data = {
    "git": git_data,
}
